@@ -39,7 +39,6 @@
 #include <linux/sched.h>
 #include <asm/segment.h>
 #include <linux/types.h>
-#include <linux/wrapper.h>
 
 #include <linux/videodev.h>
 #include <linux/version.h>
@@ -49,10 +48,14 @@
 #include <linux/i2c-algo-bit.h>
 #include <linux/video_encoder.h>
 
+#include "em8300_compat24.h"
 #include "bt865.h"
 #include "encoder.h"
 
 MODULE_SUPPORTED_DEVICE("bt865");
+MODULE_LICENSE("GPL");
+
+EXPORT_NO_SYMBOLS;
 
 static int color_bars[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 MODULE_PARM(color_bars, "1-" __MODULE_STRING(EM8300_MAX) "i");
@@ -62,12 +65,6 @@ static int rgb_mode[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 MODULE_PARM(rgb_mode, "1-" __MODULE_STRING(EM8300_MAX) "i");
 MODULE_PARM_DESC(rgb_mode, "If you set this to 1, RGB output is enabled. You will need to hack the DXR3 hardware. Defaults to 0.");
 
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
-MODULE_LICENSE("GPL");
-#endif
-
-EXPORT_NO_SYMBOLS;
-
 #define i2c_is_isa_client(clientptr) \
 		((clientptr)->adapter->algo->id == I2C_ALGO_ISA)
 #define i2c_is_isa_adapter(adapptr) \
@@ -92,7 +89,7 @@ struct bt865_data_s {
 
 /* This is the driver that will be inserted */
 static struct i2c_driver bt865_driver = {
-#ifdef _LINUX_I2C_H /* i2c version 2.8.0 or above */
+#if defined(EM8300_I2C_FORCE_NEW_API) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,54) && !defined(EM8300_I2C_FORCE_OLD_API))
 	.owner =		THIS_MODULE,
 #endif
 	.name =			"BT865 video encoder driver",
@@ -756,7 +753,7 @@ static unsigned char PALNC_CONFIG_BT865[ 48 ] = {
 
 static int bt865_update( struct i2c_client *client )
 {
-	struct bt865_data_s *data = client->data ;
+	struct bt865_data_s *data = i2c_get_clientdata(client);
 	char tmpconfig[48];
 	int i;
 
@@ -786,7 +783,7 @@ static int bt865_update( struct i2c_client *client )
 
 static int bt865_setmode(int mode, struct i2c_client *client)
 {
-	struct bt865_data_s *data = client->data ;
+	struct bt865_data_s *data = i2c_get_clientdata(client);
 	unsigned char *config = NULL;
 
 	pr_debug("bt865_setmode( %d, %p )\n", mode, client);
@@ -840,8 +837,8 @@ static int bt865_setmode(int mode, struct i2c_client *client)
 
 static int bt865_setup(struct i2c_client *client)
 {
-	struct bt865_data_s *data = client->data ;
-	struct em8300_s *em = client->adapter->data;
+	struct bt865_data_s *data = i2c_get_clientdata(client);
+	struct em8300_s *em = i2c_get_adapdata(client->adapter);
 
 	if (memset(data->config, 0, sizeof(data->config)) != data->config) {
 		printk(KERN_NOTICE "bt865_setup: memset error\n");
@@ -875,7 +872,6 @@ static int bt865_detect(struct i2c_adapter *adapter, int address)
 	int err, chk = 0;
 
 	if (i2c_is_isa_adapter(adapter)) {
-		printk(KERN_ERR "bt865a.o: called for an ISA bus adapter?!?\n");
 		return 0;
 	}
 
@@ -892,14 +888,15 @@ static int bt865_detect(struct i2c_adapter *adapter, int address)
 	if (!new_client) {
 		return -ENOMEM;
 	}
-
+	memset(new_client, 0, sizeof(struct i2c_client) + sizeof(struct bt865_data_s));
 	data = (struct bt865_data_s *) (((struct i2c_client *) new_client) + 1);
 	new_client->addr = address;
-	new_client->data = data;
 	new_client->adapter = adapter;
 	new_client->driver = &bt865_driver;
 	new_client->flags = 0;
 
+	i2c_set_clientdata(new_client, data);
+
 //	i2c_smbus_write_byte_data(new_client,0xa6, 0x80/*b1*/);
 //	the write is not needed
 //	from the left we have three bits identifying the chip
@@ -910,7 +907,9 @@ static int bt865_detect(struct i2c_adapter *adapter, int address)
 		strcpy(new_client->name, "BT865 chip");
 		printk(KERN_NOTICE "bt865.o: BT865 chip detected\n");
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
 		new_client->id = bt865_id++;
+#endif
 
 		if ((err = i2c_attach_client(new_client))) {
 			kfree(new_client);
@@ -921,9 +920,7 @@ static int bt865_detect(struct i2c_adapter *adapter, int address)
 			return -1;
 		}
 
-#ifdef MODULE
-		MOD_INC_USE_COUNT;
-#endif
+		EM8300_MOD_INC_USE_COUNT;
 
 		return 0;
 	}
@@ -948,9 +945,7 @@ int bt865_detach_client(struct i2c_client *client)
 		return err;
 	}
 
-#ifdef MODULE
-	MOD_DEC_USE_COUNT;
-#endif
+	EM8300_MOD_DEC_USE_COUNT;
 
 	kfree(client);
 
@@ -959,7 +954,7 @@ int bt865_detach_client(struct i2c_client *client)
 
 int bt865_command(struct i2c_client *client, unsigned int cmd, void *arg)
 {
-	struct bt865_data_s *data = client->data ;
+	struct bt865_data_s *data = i2c_get_clientdata(client);
 
 	switch(cmd) {
 	case ENCODER_CMD_SETMODE:
