@@ -17,7 +17,7 @@
 	along with this program; if not, write to the Free Software
 	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
-#include <linux/autoconf.h>
+#include <linux/config.h>
 #include <linux/version.h>
 #include <linux/module.h>
 #include <linux/delay.h>
@@ -35,18 +35,7 @@
 #include <linux/string.h>
 #include <linux/time.h>
 #include <linux/poll.h>
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
-#include <linux/wrapper.h>	/* for mem_map_reserve */
-#endif
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
-#include <linux/kcomp.h>
-#endif
-#ifdef CONFIG_PROC_FS
-#include <linux/proc_fs.h>
-#endif
-#ifdef CONFIG_DEVFS_FS
 #include <linux/devfs_fs_kernel.h>
-#endif
 #include <asm/io.h>
 #include <asm/pgtable.h>
 #include <asm/page.h>
@@ -56,31 +45,23 @@
 #include <asm/mtrr.h>
 #endif
 
+#include <linux/interrupt.h>
 #include <asm/uaccess.h>
 #include <linux/i2c.h>
 #include <linux/i2c-algo-bit.h>
 
+#include "em8300_compat24.h"
 #include "encoder.h"
 
 #include "em8300_reg.h"
 #include <linux/em8300.h>
 #include "em8300_fifo.h"
+#include "em8300_registration.h"
 
 #ifdef CONFIG_EM8300_IOCTL32
 #include "em8300_ioctl32.h"
 #endif
 
-/* It seems devfs will implement a new scheme of enumerating minor numbers.
- * Currently it seems broken. But that is why we added these macros.
- */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-#define EM8300_MINOR(inode) (MINOR((inode)->i_rdev) % 4)
-#define EM8300_CARD(inode) (MINOR((inode)->i_rdev) / 4)
-#else
-#define EM8300_MINOR(inode) (minor(inode->i_rdev) % 4)
-#define EM8300_CARD(inode) (minor(inode->i_rdev) / 4)
-#endif
-
 #if !defined(CONFIG_I2C_ALGOBIT) && !defined(CONFIG_I2C_ALGOBIT_MODULE)
 #error "This needs the I2C Bit Banging Interface in your Kernel"
 #endif
@@ -88,13 +69,9 @@
 MODULE_AUTHOR("Henrik Johansson <henrikjo@post.utfors.se>");
 MODULE_DESCRIPTION("EM8300 MPEG-2 decoder");
 MODULE_SUPPORTED_DEVICE("em8300");
-
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
-MODULE_PARM(remap, "1-" __MODULE_STRING(EM8300_MAX) "i");
-#endif
-
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
 MODULE_LICENSE("GPL");
+#if EM8300_MAJOR != 0
+MODULE_ALIAS_CHARDEV_MAJOR(EM8300_MAJOR);
 #endif
 
 EXPORT_NO_SYMBOLS;
@@ -146,23 +123,24 @@ int activate_loopback[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 MODULE_PARM(activate_loopback, "1-" __MODULE_STRING(EM8300_MAX) "i");
 MODULE_PARM_DESC(activate_loopback, "If you lose video after loading the modules or uploading the microcode set this to 1. Defaults to 0.");
 
+int major = EM8300_MAJOR;
+MODULE_PARM(major, "i");
+MODULE_PARM_DESC(major, "Major number used for the devices. "
+		 "0 means automatically assigned. "
+		 "Defaults to " __MODULE_STRING(EM8300_MAJOR) ".");
+
 static int em8300_cards,clients;
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
-static unsigned int remap[EM8300_MAX]={};
-#endif
 static struct em8300_s em8300[EM8300_MAX];
+
 #if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
+int dsp_num[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = -1 };
+MODULE_PARM(dsp_num, "1-" __MODULE_STRING(EM8300_MAX) "i");
+MODULE_PARM_DESC(dsp_num, "The /dev/dsp number to assign to the card. -1 for automatic (this is the default).");
+
 static int dsp_num_table[16];
 #endif
-#ifdef CONFIG_DEVFS_FS
-devfs_handle_t em8300_handle[EM8300_MAX*4];
-#endif
-#ifdef CONFIG_PROC_FS
-struct proc_dir_entry *em8300_proc;
-#endif
 
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
 /* structure to keep track of the memory that has been allocated by
    the user via mmap() */
 struct memory_info
@@ -171,9 +149,16 @@ struct memory_info
 	long length;
 	char *ptr;
 };
-#endif
 
-static void em8300_irq(int irq, void *dev_id, struct pt_regs * regs)
+static struct pci_device_id em8300_ids[] = {
+	{ PCI_VENDOR_ID_SIGMADESIGNS, PCI_DEVICE_ID_SIGMADESIGNS_EM8300,
+	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
+	{ 0 }
+};
+
+MODULE_DEVICE_TABLE(pci, em8300_ids);
+
+static irqreturn_t em8300_irq(int irq, void *dev_id, struct pt_regs * regs)
 {
 	struct em8300_s *em = (struct em8300_s *) dev_id;
 	int irqstatus;
@@ -209,121 +194,46 @@ static void em8300_irq(int irq, void *dev_id, struct pt_regs * regs)
 
 		write_ucregister(Q_IrqMask, em->irqmask);
 		write_ucregister(Q_IrqStatus, 0x0000);
+		return IRQ_HANDLED;
 	}
+	return IRQ_NONE;
 }
 
-static void release_em8300(int max)
+static void release_em8300(struct em8300_s *em)
 {
-	struct em8300_s *em;
-	int i;
-
-	for (i = 0; i < max; i++) {
-		em = &em8300[i];
-
-		if(em->encoder) {
-			em->encoder->driver->command(em->encoder, ENCODER_CMD_ENABLEOUTPUT, (void *) 0);
-		}
+	if(em->encoder) {
+		em->encoder->driver->command(em->encoder, ENCODER_CMD_ENABLEOUTPUT, (void *) 0);
+	}
 
 #ifdef CONFIG_MTRR
-		if (em->mtrr_reg) {
-			mtrr_del(em->mtrr_reg,em->adr, em->memsize);
-		}
-#endif
-
-		em8300_i2c_exit(em);
-
-		write_ucregister(Q_IrqMask, 0);
-		write_ucregister(Q_IrqStatus, 0);
-		writel(0, &em->mem[0x2000]);
-
-		em8300_fifo_free(em->mvfifo);
-		em8300_fifo_free(em->mafifo);
-		em8300_fifo_free(em->spfifo);
-
-		/* free it */
-		free_irq(em->dev->irq, em);
-
-		/* unmap and free memory */
-		if (em->mem) {
-			iounmap((unsigned *) em->mem);
-		}
+	if (em->mtrr_reg) {
+		mtrr_del(em->mtrr_reg,em->adr, em->memsize);
 	}
-}
-
-static int find_em8300(void)
-{
-	struct pci_dev *dev = NULL;
-	unsigned char revision;
-	struct em8300_s *em;
-	unsigned int em8300_n = 0;
-	int result;
-
-	while ((dev = pci_find_device(PCI_VENDOR_ID_SIGMADESIGNS, PCI_DEVICE_ID_SIGMADESIGNS_EM8300, dev))) {
-		em = &em8300[em8300_n];
-		em->dev = dev;
-		em->card_nr = em8300_n;
-
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
-		em->adr = dev->base_address[0];
-		if (remap[em8300_n]) {
-			uint dw;
-
-			if (remap[em8300_n] < 0x1000) {
-				remap[em8300_n] <<= 20;
-			}
-			remap[em8300_n] &= PCI_BASE_ADDRESS_MEM_MASK;
-			pr_info("Remapping to : 0x%08x.\n", remap[em8300_n]);
-			remap[em8300_n] |= em->adr&(~PCI_BASE_ADDRESS_MEM_MASK);
-			pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, remap[em8300_n]);
-			/* commit to PCI bus */
-			pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &dw);
-			dev->base_address[0] = em->adr;
-		}
-		em->adr &= PCI_BASE_ADDRESS_MEM_MASK;
-#else
-		em->adr = dev->resource[0].start;
-#endif
-		em->memsize = 1024 * 1024;
-
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
-		pci_enable_device(dev);
-#endif
-		pci_read_config_byte(dev, PCI_CLASS_REVISION, &revision);
-		em->pci_revision = revision;
-		pr_info("em8300: EM8300 %x (rev %d) ", dev->device, revision);
-		printk("bus: %d, devfn: %d, irq: %d, ", dev->bus->number, dev->devfn, dev->irq);
-		printk("memory: 0x%08lx.\n", em->adr);
-
-		em->mem = ioremap(em->adr, em->memsize);
-		pr_info("em8300: mapped-memory at 0x%p\n", em->mem);
-#ifdef CONFIG_MTRR
-		em->mtrr_reg = mtrr_add(em->adr, em->memsize, MTRR_TYPE_UNCACHABLE, 1);
-		if (em->mtrr_reg) pr_info("em8300: using MTRR\n");
 #endif
 
-		result = request_irq(dev->irq, em8300_irq, SA_SHIRQ | SA_INTERRUPT, "em8300", (void *) em);
+	em8300_i2c_exit(em);
 
-		if (result == -EINVAL) {
-			printk(KERN_ERR "em8300: Bad irq number or handler\n");
-			return -EINVAL;
-		}
+	write_ucregister(Q_IrqMask, 0);
+	write_ucregister(Q_IrqStatus, 0);
+	writel(0, &em->mem[0x2000]);
 
-		pci_set_master(dev);
+	em8300_fifo_free(em->mvfifo);
+	em8300_fifo_free(em->mafifo);
+	em8300_fifo_free(em->spfifo);
 
-		em8300_n++;
+	/* free it */
+	free_irq(em->dev->irq, em);
+
+	/* unmap and free memory */
+	if (em->mem) {
+		iounmap((unsigned *) em->mem);
 	}
-
-	if (em8300_n) {
-		pr_info("em8300: %d EM8300 card(s) found.\n", em8300_n);
-	}
-
-	return em8300_n;
 }
 
 static int em8300_io_ioctl(struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg)
 {
 	struct em8300_s *em = filp->private_data;
-	int subdevice = EM8300_MINOR(inode);
+	int subdevice = EM8300_IMINOR(inode) % 4;
 
 	switch (subdevice) {
 	case EM8300_SUBDEVICE_AUDIO:
@@ -341,8 +251,8 @@ static int em8300_io_ioctl(struct inode* inode, struct file* filp, unsigned int
 
 static int em8300_io_open(struct inode* inode, struct file* filp)
 {
-	int card = EM8300_CARD(inode);
-	int subdevice = EM8300_MINOR(inode);
+	int card = EM8300_IMINOR(inode) / 4;
+	int subdevice = EM8300_IMINOR(inode) % 4;
 	struct em8300_s *em = &em8300[card];
 	int err = 0;
 
@@ -358,10 +268,9 @@ static int em8300_io_open(struct inode* inode, struct file* filp)
 
 	filp->private_data = &em8300[card];
 
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
 	/* initalize the memory list */
 	INIT_LIST_HEAD(&em->memory);
-#endif
+
 	switch (subdevice) {
 	case EM8300_SUBDEVICE_CONTROL:
 		em8300[card].nonblock[0] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
@@ -401,7 +310,8 @@ static int em8300_io_open(struct inode* inode, struct file* filp)
 
 	clients++;
 	pr_debug("em8300_main.o: Opening device %d, Clients:%d\n", subdevice, clients);
-	MOD_INC_USE_COUNT;
+
+	EM8300_MOD_INC_USE_COUNT;
 
 	return(0);
 }
@@ -409,7 +319,7 @@ static int em8300_io_open(struct inode* inode, struct file* filp)
 static ssize_t em8300_io_write(struct file *file, const char * buf, size_t count, loff_t *ppos)
 {
 	struct em8300_s *em = file->private_data;
-	int subdevice = EM8300_MINOR(file->f_dentry->d_inode);
+	int subdevice = EM8300_IMINOR(file->f_dentry->d_inode) % 4;
 
 	switch (subdevice) {
 	case EM8300_SUBDEVICE_VIDEO:
@@ -433,18 +343,13 @@ int em8300_io_mmap(struct file *file, struct vm_area_struct *vma)
 {
 	struct em8300_s *em = file->private_data;
 	unsigned long size = vma->vm_end - vma->vm_start;
-	int subdevice = EM8300_MINOR(file->f_dentry->d_inode);
+	int subdevice = EM8300_IMINOR(file->f_dentry->d_inode) % 4;
 
 	if (subdevice != EM8300_SUBDEVICE_CONTROL) {
 		return -EPERM;
 	}
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
-	switch ((unsigned long) vma->vm_offset) {
-#else
 	switch (vma->vm_pgoff) {
-#endif
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
 	case 1: {
 		/* fixme: we should count the total size of allocated memories
 		   so we don't risk a out-of-memory or denial-of-service attack... */
@@ -467,14 +372,20 @@ int em8300_io_mmap(struct file *file, struct vm_area_struct *vma)
 
 		/* reserve all pages */
 		for(adr = (long)mem; adr < (long)mem + size; adr += PAGE_SIZE) {
-			mem_map_reserve(virt_to_page(adr));
+			SetPageReserved(virt_to_page(adr));
 		}
 
 		/* lock the area*/
 		vma->vm_flags |=VM_LOCKED;
 
 		/* remap the memory to user space */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,3)
 		if (remap_page_range(vma->vm_start, virt_to_phys((void *)mem), size, vma->vm_page_prot)) {
+#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
+		if (remap_page_range(vma, vma->vm_start, virt_to_phys((void *)mem), size, vma->vm_page_prot)) {
+#else
+		if (remap_pfn_range(vma, vma->vm_start, virt_to_phys((void *)mem) >> PAGE_SHIFT, size, vma->vm_page_prot)) {
+#endif
 			kfree(mem);
 			return -EAGAIN;
 		}
@@ -495,24 +406,20 @@ int em8300_io_mmap(struct file *file, struct vm_area_struct *vma)
 
 		break;
 	}
-#endif
 	case 0:
 		if (size > em->memsize) {
 			return -EINVAL;
 		}
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,3)
 		remap_page_range(vma->vm_start, em->adr, vma->vm_end - vma->vm_start, vma->vm_page_prot);
-#else
+#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 		remap_page_range(vma, vma->vm_start, em->adr, vma->vm_end - vma->vm_start, vma->vm_page_prot);
+#else
+		remap_pfn_range(vma, vma->vm_start, em->adr >> PAGE_SHIFT, vma->vm_end - vma->vm_start, vma->vm_page_prot);
 #endif
 		vma->vm_file = file;
-
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
-		file->f_dentry->d_inode->i_count++;
-#else
 		atomic_inc(&file->f_dentry->d_inode->i_count);
-#endif
 		break;
 	default:
 		return -EINVAL;
@@ -524,7 +431,7 @@ int em8300_io_mmap(struct file *file, struct vm_area_struct *vma)
 static unsigned int em8300_poll(struct file *file, struct poll_table_struct *wait)
 {
 	struct em8300_s *em = file->private_data;
-	int subdevice = EM8300_MINOR(file->f_dentry->d_inode);
+	int subdevice = EM8300_IMINOR(file->f_dentry->d_inode) % 4;
 	unsigned int mask = 0;
 
 	switch (subdevice) {
@@ -562,7 +469,7 @@ static unsigned int em8300_poll(struct file *file, struct poll_table_struct *wai
 int em8300_io_release(struct inode* inode, struct file *filp)
 {
 	struct em8300_s *em = filp->private_data;
-	int subdevice = EM8300_MINOR(inode);
+	int subdevice = EM8300_IMINOR(inode) % 4;
 
 	switch (subdevice) {
 	case EM8300_SUBDEVICE_AUDIO:
@@ -573,10 +480,10 @@ int em8300_io_release(struct inode* inode, struct file *filp)
 		em8300_ioctl_enable_videoout(em, 0);
 		break;
 	case EM8300_SUBDEVICE_SUBPICTURE:
+		em8300_spu_release(em);    /* Do we need this one ? */
 		break;
 	}
 
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
 	while( 0 == list_empty(&em->memory)) {
 		unsigned long adr = 0;
 
@@ -584,24 +491,25 @@ int em8300_io_release(struct inode* inode, struct file *filp)
 		list_del(&info->item);
 
 		for(adr = (long)info->ptr; adr < (long)info->ptr + info->length; adr += PAGE_SIZE) {
-			mem_map_unreserve(virt_to_page(adr));
+			ClearPageReserved(virt_to_page(adr));
 		}
 
 		kfree(info->ptr);
 		vfree(info);
 	}
-#endif
 
 	em->inuse[subdevice]--;
 
 	clients--;
 	pr_debug("em8300_main.o: Releasing device %d, clients:%d\n", subdevice, clients);
-	MOD_DEC_USE_COUNT;
+
+	EM8300_MOD_DEC_USE_COUNT;
 
 	return(0);
 }
 
-static struct file_operations em8300_fops = {
+struct file_operations em8300_fops = {
+	owner: THIS_MODULE,
 	write: em8300_io_write,
 	ioctl: em8300_io_ioctl,
 	mmap: em8300_io_mmap,
@@ -619,15 +527,11 @@ static int em8300_dsp_ioctl(struct inode* inode, struct file* filp, unsigned int
 
 static int em8300_dsp_open(struct inode* inode, struct file* filp)
 {
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-	int dsp_num = ((MINOR(inode->i_rdev) >> 4) & 0x0f);
-#else
-	int dsp_num = ((minor(inode->i_rdev) >> 4) & 0x0f);
-#endif
-	int card = dsp_num_table[dsp_num] - 1;
+	int dsp_number = ((EM8300_IMINOR(inode) >> 4) & 0x0f);
+	int card = dsp_num_table[dsp_number] - 1;
 	int err = 0;
 
-	pr_debug("em8300: opening dsp %i for card %i\n", dsp_num, card);
+	pr_debug("em8300: opening dsp %i for card %i\n", dsp_number, card);
 
 	if (card < 0 || card >= em8300_cards) {
 		return -ENODEV;
@@ -649,7 +553,8 @@ static int em8300_dsp_open(struct inode* inode, struct file* filp)
 
 	clients++;
 	pr_debug("em8300_main.o: Opening device %d, Clients:%d\n", EM8300_SUBDEVICE_AUDIO, clients);
-	MOD_INC_USE_COUNT;
+
+	EM8300_MOD_INC_USE_COUNT;
 
 	return(0);
 }
@@ -684,12 +589,14 @@ int em8300_dsp_release(struct inode* inode, struct file* filp)
 
 	clients--;
 	pr_debug("em8300_main.o: Releasing device %d, clients:%d\n", EM8300_SUBDEVICE_AUDIO, clients);
-	MOD_DEC_USE_COUNT;
+
+	EM8300_MOD_DEC_USE_COUNT;
 
 	return(0);
 }
 
 static struct file_operations em8300_dsp_audio_fops = {
+	owner: THIS_MODULE,
 	write: em8300_dsp_write,
 	ioctl: em8300_dsp_ioctl,
 	poll: em8300_dsp_poll,
@@ -698,45 +605,7 @@ static struct file_operations em8300_dsp_audio_fops = {
 };
 #endif
 
-#ifdef CONFIG_PROC_FS
-int em8300_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
-{
-	int len = 0;
-	struct em8300_s *em = (struct em8300_s *) data;
-
-	*start = 0;
-	*eof = 1;
-
-	len += sprintf(page + len, "----- Driver Info -----\n");
-	len += sprintf(page + len, "em8300 module version %s\n", EM8300_VERSION);
-	if (em->ucodeloaded) {
-		/* Device information */
-		len += sprintf(page + len, "Card revision %d\n", em->pci_revision);
-		len += sprintf(page + len, "Chip revision %d\n", em->chip_revision);
-		len += sprintf(page + len, "Video encoder: %s at address 0x%02x on %s\n", (em->encoder_type == ENCODER_BT865) ? "BT865" : (em->encoder_type == ENCODER_ADV7170) ? "ADV7170" : (em->encoder_type == ENCODER_ADV7175) ? "ADV7175" : "unknown", em->encoder->addr, em->encoder->adapter->name);
-		len += sprintf(page + len, "Memory mapped at addressrange 0x%0lx->0x%0lx%s\n", (unsigned long int) em->mem,
-				(unsigned long int) em->mem + (unsigned long int) em->memsize, (em->mtrr_reg ? " (FIFOs using MTRR)" : ""));
-		len += sprintf(page + len, "Displaybuffer resolution: %dx%d\n", em->dbuf_info.xsize, em->dbuf_info.ysize);
-		len += sprintf(page + len, "Dicom set to %s\n", (em->dicom_tvout?"TV-out":"overlay"));
-		if (em->dicom_tvout) {
-			len += sprintf(page + len, "Using %s\n", (em->video_mode == EM8300_VIDEOMODE_PAL ? "PAL" : "NTSC"));
-			len += sprintf(page + len, "Aspect is %s\n", (em->aspect_ratio == EM8300_ASPECTRATIO_4_3 ? "4:3" : "16:9"));
-		} else {
-			len += sprintf(page + len, "em9010 %s\n", (em->overlay_enabled ? "online" : "offline"));
-			len += sprintf(page + len, "video mapped to screen coordinates %dx%d (%dx%d)\n", em->overlay_frame_xpos,
-					em->overlay_frame_ypos, em->overlay_xres, em->overlay_yres);
-		}
-		len += sprintf(page + len, "%s audio output\n", (em->audio_mode == EM8300_AUDIOMODE_ANALOG ? "analog" : "digital"));
-	}
-	else {
-		len += sprintf(page + len, "Microcode hasn't been loaded\n");
-	}
-
-	return len;
-}
-#endif
-
-int init_em8300(struct em8300_s *em)
+static int init_em8300(struct em8300_s *em)
 {
 	write_register(0x30000, read_register(0x30000));
 
@@ -788,125 +657,147 @@ int init_em8300(struct em8300_s *em)
 	return 0;
 }
 
-void __exit em8300_exit(void)
+static int __devinit em8300_probe(struct pci_dev *dev,
+				  const struct pci_device_id *pci_id)
 {
-	int card;
-#ifdef CONFIG_DEVFS_FS
-	int frame;
-#endif
-	char devname[64];
+	unsigned char revision;
+	struct em8300_s *em;
+	int result;
 
+	em = &em8300[em8300_cards];
+	em->dev = dev;
+	em->card_nr = em8300_cards;
+	em->adr = dev->resource[0].start;
+	em->memsize = 1024 * 1024;
+
+	if ((result = pci_enable_device(dev)) != 0) {
+		printk(KERN_ERR "em8300: Unable to enable PCI device\n");
+		return result;
+	}
+
+	pci_read_config_byte(dev, PCI_CLASS_REVISION, &revision);
+	em->pci_revision = revision;
+	pr_info("em8300: EM8300 %x (rev %d) ", dev->device, revision);
+	printk("bus: %d, devfn: %d, irq: %d, ", dev->bus->number, dev->devfn, dev->irq);
+	printk("memory: 0x%08lx.\n", em->adr);
+
+	em->mem = ioremap(em->adr, em->memsize);
+	pr_info("em8300: mapped-memory at 0x%p\n", em->mem);
+#ifdef CONFIG_MTRR
+	em->mtrr_reg = mtrr_add(em->adr, em->memsize, MTRR_TYPE_UNCACHABLE, 1);
+	if (em->mtrr_reg) pr_info("em8300: using MTRR\n");
+#endif
+
+	init_waitqueue_head(&em->video_ptsfifo_wait);
+	init_waitqueue_head(&em->vbi_wait);
+	init_waitqueue_head(&em->sp_ptsfifo_wait);
+
+	result = request_irq(dev->irq, em8300_irq, SA_SHIRQ | SA_INTERRUPT, "em8300", (void *) em);
+
+	if (result == -EINVAL) {
+		printk(KERN_ERR "em8300: Bad irq number or handler\n");
+		return -EINVAL;
+	}
+
+	pci_set_master(dev);
+	pci_set_drvdata(dev, em);
+
+	em->irqmask = 0;
+	em->encoder = NULL;
+	em->linecounter=0;
+
+	init_em8300(em);
+
+	em8300_register_card(em);
+
+#if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
+	if ((em->dsp_num = register_sound_dsp(&em8300_dsp_audio_fops, dsp_num[em->card_nr])) < 0) {
+		printk(KERN_ERR "em8300: cannot register oss audio device!\n");
+	} else {
+		dsp_num_table[em->dsp_num >> 4 & 0x0f] = em8300_cards + 1;
+		pr_debug("em8300: registered dsp %i for device %i\n", em->dsp_num >> 4 & 0x0f, em8300_cards);
+	}
+#endif
+
+	em8300_cards++;
+	return 0;
+}
+
+static void __devexit em8300_remove(struct pci_dev *pci)
+{
+	struct em8300_s *em = pci_get_drvdata(pci);
+
+	if (em) {
+		if (em->ucodeloaded == 1)
+			em8300_disable_card(em);
+
+#if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
+		unregister_sound_dsp(em->dsp_num);
+#endif
+		em8300_unregister_card(em);
+
+		release_em8300(em);
+	}
+
+	pci_set_drvdata(pci, NULL);
+	pci_disable_device(pci);
+}
+
+static struct pci_driver em8300_driver = {
+	.name     = "Sigma Designs EM8300",
+	.id_table = em8300_ids,
+	.probe    = em8300_probe,
+	.remove   = __devexit_p(em8300_remove),
+};
+
+static void __exit em8300_exit(void)
+{
 #ifdef CONFIG_EM8300_IOCTL32
 	em8300_ioctl32_exit();
 #endif
 
-	for (card = 0; card < em8300_cards; card++) {
-#ifdef CONFIG_DEVFS_FS
-		for (frame = 3; frame >= 0; frame--) {
-			devfs_unregister(em8300_handle[(card * 4) + frame]);
-		}
-#endif
-#if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
-		unregister_sound_dsp(em8300[card].dsp_num);
-#endif
-	}
-#ifdef CONFIG_PROC_FS
-	sprintf(devname, "%s", EM8300_LOGNAME );
-	if (em8300_proc != NULL) remove_proc_entry(devname, &proc_root);
-#endif
-#if defined(CONFIG_DEVFS_FS) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-	sprintf(devname, "%s", EM8300_LOGNAME);
-	devfs_unregister_chrdev(EM8300_MAJOR, devname);
-#endif
-	unregister_chrdev(EM8300_MAJOR, EM8300_LOGNAME);
-	release_em8300(em8300_cards);
+	pci_unregister_driver(&em8300_driver);
+
+	unregister_chrdev(major, EM8300_LOGNAME);
+
+	em8300_unregister_driver();
 }
 
-int __init em8300_init(void)
+static int __init em8300_init(void)
 {
-	int card = 0;
-	int frame = 3;
-	struct em8300_s *em = NULL;
+	int err;
 
-	char devname[32];
-#ifdef CONFIG_PROC_FS
-	struct proc_dir_entry *proc;
-#endif
 	//memset(&em8300, 0, sizeof(em8300) * EM8300_MAX);
 #if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
 	memset(&dsp_num_table, 0, sizeof(dsp_num_table));
 #endif
 
-#if defined(CONFIG_DEVFS_FS) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-	sprintf(devname, "%s", EM8300_LOGNAME);
-	devfs_register_chrdev(EM8300_MAJOR, devname, &em8300_fops);
-#endif
-#ifdef CONFIG_PROC_FS
-	sprintf(devname, "%s", EM8300_LOGNAME);
-	em8300_proc = create_proc_entry(devname, S_IFDIR | S_IRUGO | S_IXUGO, &proc_root);
-	if (em8300_proc == NULL) {
-		printk(KERN_ERR "em8300: unable to register proc entry!\n");
-		goto err_chrdev;
-	}
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
-	em8300_proc->owner = THIS_MODULE;
-#endif
-#endif
-#ifdef CONFIG_SOUND_MODULE
-	//request_module("soundcore");
-#endif
+	em8300_register_driver();
 
-	/* Find EM8300 cards */
-	em8300_cards = find_em8300();
-
-	/* Initialize EM8300 cards */
-	for (card = 0; card < em8300_cards; card++) {
-		em = &em8300[card];
-
-		em->irqmask = 0;
-
-		em->encoder = NULL;
-
-		em->linecounter=0;
-
-		init_em8300(em);
-
-#ifdef CONFIG_PROC_FS
-		sprintf(devname, "%d", card );
-		proc = create_proc_entry(devname, S_IFREG | S_IRUGO, em8300_proc);
-		proc->data = (void *) em;
-		proc->read_proc = em8300_proc_read;
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
-		proc->owner = THIS_MODULE;
-#endif
-#endif
-#ifdef CONFIG_DEVFS_FS
-		sprintf(devname, "%s-%d", EM8300_LOGNAME, card );
-		em8300_handle[(card * 4)] = devfs_register(NULL, devname, DEVFS_FL_DEFAULT, EM8300_MAJOR,
-				(card * 4), S_IFCHR | S_IRUGO | S_IWUGO, &em8300_fops, NULL);
-		sprintf(devname, "%s_mv-%d", EM8300_LOGNAME, card );
-		em8300_handle[(card * 4) + 1] = devfs_register(NULL, devname, DEVFS_FL_DEFAULT, EM8300_MAJOR,
-				(card * 4)+1, S_IFCHR | S_IRUGO | S_IWUGO, &em8300_fops, NULL);
-		sprintf(devname, "%s_ma-%d", EM8300_LOGNAME, card );
-		em8300_handle[(card * 4) + 2] = devfs_register(NULL, devname, DEVFS_FL_DEFAULT, EM8300_MAJOR,
-				(card * 4) + 2, S_IFCHR | S_IRUGO | S_IWUGO, &em8300_fops, NULL);
-		sprintf(devname, "%s_sp-%d", EM8300_LOGNAME, card );
-		em8300_handle[(card * 4) + 3] = devfs_register(NULL, devname, DEVFS_FL_DEFAULT, EM8300_MAJOR,
-				(card * 4) + 3, S_IFCHR | S_IRUGO | S_IWUGO, &em8300_fops, NULL);
-#endif
-#if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
-		if ((em8300[card].dsp_num = register_sound_dsp(&em8300_dsp_audio_fops, -1)) < 0) {
-			printk(KERN_ERR "em8300: cannot register oss audio device!\n");
-			goto err_audio_dsp;
+	if (major) {
+		if (register_chrdev(major, EM8300_LOGNAME, &em8300_fops)) {
+			printk(KERN_ERR "em8300: unable to get major %d\n", major);
+			err = -ENODEV;
+			goto err_chrdev;
+		}
+	}
+	else {
+		int m = register_chrdev(major, EM8300_LOGNAME, &em8300_fops);
+		if (m > 0) {
+			major = m;
+		}
+		else {
+			printk(KERN_ERR "em8300: unable to get any majo\n");
+			err = -ENODEV;
+			goto err_chrdev;
 		}
-		dsp_num_table[em8300[card].dsp_num >> 4 & 0x0f] = card + 1;
-		pr_debug("em8300: registered dsp %i for device %i\n", em8300[card].dsp_num >> 4 & 0x0f, card);
-#endif
 	}
 
-	if (register_chrdev(EM8300_MAJOR, EM8300_LOGNAME, &em8300_fops)) {
-		printk(KERN_ERR "em8300: unable to get major %d\n", EM8300_MAJOR);
-		goto err_chrdev;
+	if ((err = pci_module_init(&em8300_driver)) < 0) {
+#ifdef MODULE
+		printk(KERN_ERR "Sigmadesigns EM8300 not found or device busy\n");
+#endif
+		goto err_init;
 	}
 
 #ifdef CONFIG_EM8300_IOCTL32
@@ -915,27 +806,12 @@ int __init em8300_init(void)
 
 	return 0;
 
-#if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
- err_audio_dsp:
-#endif
+ err_init:
+	unregister_chrdev(major, EM8300_LOGNAME);
+
  err_chrdev:
-	while (card-- > 0) {
-		while (frame-- > 0) {
-#ifdef CONFIG_DEVFS_FS
-			devfs_unregister(em8300_handle[(card * 4) + frame]);
-#endif
-		}
-		frame = 3;
-#if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
-		unregister_sound_dsp(em[card].dsp_num);
-#endif
-	}
-#if defined(CONFIG_DEVFS_FS) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-	sprintf(devname, "%s", EM8300_LOGNAME);
-	devfs_unregister_chrdev(EM8300_MAJOR, devname);
-#endif
-	release_em8300(em8300_cards);
-	return -ENODEV;
+	em8300_unregister_driver();
+	return err;
 }
 
 module_init(em8300_init);
